# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

import asyncio
from time import time

from . import CC_MSG,CC_DICT,CC_DATA
from ..util import uuidstr
#from aioamqp.properties import Properties
from qbroker.util import attrdict; Properties = attrdict

class _NOTGIVEN:
	pass

_types = {}
_fmap = {
	}
def fmap(s):
	r = _fmap.get(s,_NOTGIVEN)
	if r is _NOTGIVEN:
		_fmap[s] = r = s.replace('-','_')
	return r

class FieldCollect(type):
	"""\
		A metaclass which coalesces "fields" class attributes from the base
		classes into one coherent set
		"""
	def __new__(meta, name, bases, dct):
		# Grab all known field names
		s = set()
		for b in bases:
			s.update(getattr(b,'fields',()))
		b = dct.get('fields',"")
		if b:
			if isinstance(b,str):
				b = b.split(" ")
			assert not s.intersection(b), (s,b)
			s.update(b)
		dct['fields'] = s

		res = super(FieldCollect, meta).__new__(meta, name, bases, dct)
		t = dct.get('type',None)
		if t is not None:
			_types[t] = res
		return res

class _MsgPart(object, metaclass=FieldCollect):
	"""\
		A message part.
		"""
	def dump(self):
		"""Convert myself to a dict"""
		obj = {}
		for f in self.fields:
			try:
				obj[f] = getattr(self, fmap(f))
			except AttributeError:
				pass
		return obj
	
	def _load(self, props):
		"""Load myself from a proplist"""
		for f in self.fields:
			v = props.headers.get(f,_NOTGIVEN)
			if v is not _NOTGIVEN:
				setattr(self, fmap(f), v)

	def __eq__(self, other):
		for f in "type version data error".split(): # self.fields:
			a = getattr(self, f, _NOTGIVEN)
			b = getattr(other, f, _NOTGIVEN)
			if a == b:
				continue
			return False # pragma: no cover
		return True

class MsgError(RuntimeError,_MsgPart):
	"""Proxy for a remote error"""
	fields = "status id part message cls"

	def __init__(self, data=None):
		if data is not None:
			for f,v in data.items():
				setattr(self,fmap(f),v)

	@property
	def failed(self):
		"""Is this message really an error?"""
		if self.status in ('ok','warn'): # pragma: no cover ## XXX TODO
			return False
		if self.status in ('error','fail'):
			return True
		raise RuntimeError("Unknown error status: "+str(self.status)) # pragma: no cover
	
	@classmethod
	def build(cls, exc, eid,part, fail=False):
		obj = cls()
		obj.status = "fail" if fail else "error"
		obj.eid = eid
		obj.part = part
		obj.cls = exc.__class__.__name__
		obj.message = str(exc)
		return obj

	def __repr__(self):
		return "%s(%s)" % (self.cls, repr(self.message))
	def __str__(self):
		return self.message
	def __hash__(self):
		return id(self)

class BaseMsg(_MsgPart):
	version = 1
	debug = False
	# type: needs to be overridden
	fields = "type version debug message-id"
	_timer = None

	data = None
	error = None

	def __init__(self, data=None):
		if not hasattr(self,'message_id'):
			self.message_id = uuidstr()

	def __repr__(self): # pragma: no cover
		return "%s._load(%s)" % (self.__class__.__name__, repr(self.__dict__))

	def dump(self,conn):
		props = Properties()
		obj = super().dump()
		for f in 'type message-id reply-to correlation-id'.split(' '):
			m = obj.pop(f,None)
			if m is not None:
				setattr(props,fmap(f), m)
		props.timestamp = int(time())
		props.user_id = conn.cfg['login']
		props.content_type = conn.codec.CODEC
		props.app_id = conn.unit().uuid
		if self.error is not None:
			obj['error'] = self.error.dump()
		if obj:
			props.headers = obj

		data = self.data
		if data is None:
			data = ""
		return data,props

	def set_error(self, *a, **k):
		self.error = MsgError.build(*a,**k)

	@staticmethod
	def load(data,env,props):
		t = props.type
		res = _types[t]._load(data,props)

		for f in 'type message-id reply-to user-id timestamp content-type app-id correlation-id'.split(' '):
			ff = fmap(f)
			m = getattr(props,ff,_NOTGIVEN)
			if m is not _NOTGIVEN:
				setattr(res,ff,m)
		rk = getattr(res,'routing_key',None)
		if not rk:
			res.routing_key = env.routing_key
		return res

	@classmethod
	def _load(cls, msg,props):
		obj = cls()
		super(BaseMsg,obj)._load(props)
		obj.data = msg
		if 'error' in props.headers:
			obj.error = MsgError(props.headers['error'])
		return obj

	@property
	def failed(self):
		return self.error is not None and self.error.failed

	def raise_if_error(self):
		if self.error and self.error.failed:
			raise self.error

class _RequestMsg(BaseMsg):
	"""A request packet. The remaining fields are data elements."""
	fields = "routing-key reply-to"

	def __init__(self, routing_key=None, data=None):
		super().__init__()
		self.routing_key = routing_key
		self.data = data

	def make_response(self, **data):
		return ResponseMsg(self, **data)

#	def make_error_response(self, exc, eid,part, fail=False):
#		res = ResponseMsg(self)
#		res.error = MsgError.build(exc, eid,part)
#		return error

class RequestMsg(_RequestMsg):
	type = "request"
	_exchange = "rpc" # lookup key for the exchange name
	_timer = "rpc" # lookup key for the timeout

	def __init__(self, routing_key=None, _unit=None, data=None):
		super().__init__(routing_key=routing_key, data=data)
		if _unit is not None:
			self.reply_to = _unit.uuid

	@asyncio.coroutine
	def recv_reply(self, f,reply):
		"""Client side: Incoming reply. @f is the future to trigger when complete."""
		f.set_result(reply)

class AlertMsg(_RequestMsg):
	"""An alert which is not replied to"""
	type = "alert"
	_exchange = "alert" # lookup key for the exchange name

	def __init__(self, routing_key=None, _unit=None, data=None):
		super().__init__(routing_key=routing_key, data=data)
		# do not set reply_to

class PollMsg(AlertMsg):
	"""An alert which requests replies"""
	_timer = "poll" # lookup key for the timeout

	def __init__(self, routing_key=None, _unit=None, callback=None,call_conv=CC_MSG, data=None):
		super().__init__(routing_key=routing_key, _unit=_unit, data=data)
		if _unit is not None:
			self.reply_to = _unit.uuid
		self.callback = callback
		self.call_conv = call_conv
		self.replies = 0

	@asyncio.coroutine
	def recv_reply(self, f,msg):
		"""Incoming reply. @f is the future to trigger when complete."""
		try:
			if self.call_conv == CC_MSG:
				a=(msg,); k={}
			elif msg.failed:
				return # ignore error replies
			elif self.call_conv == CC_DICT:
				a=(); k=msg.data
			elif self.call_conv == CC_DATA:
				a=(msg.data,); k={}
			else: # pragma: no cover
				raise RuntimeError("Unknown encoding: %s"%self.call_conv) 
			r = self.callback(*a,**k)
			if asyncio.iscoroutine(r):
				yield from r
		except StopIteration:
			f.set_result(self.replies+1)
		except Exception as exc:
			f.set_exception(exc)
		else:
			self.replies += 1

class ResponseMsg(BaseMsg):
	type = "reply"
	fields = "correlation-id"

	def __init__(self,request=None, data=None):
		super().__init__(data=data)
		if request is not None:
			self.correlation_id = request.message_id




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

import asyncio
import inspect

from . import CC_MSG,CC_DICT,CC_DATA
from ..util import attrdict, import_string, uuidstr

if not hasattr(inspect,'iscoroutinefunction'):
	inspect.iscoroutinefunction = inspect.isgeneratorfunction
if not hasattr(inspect,'iscoroutine'):
	inspect.iscoroutine = inspect.isgenerator
if not hasattr(inspect,'isawaitable'):
	def isawaitable(object):
		return False
	inspect.isawaitable = isawaitable

@asyncio.coroutine
def coro_wrapper(proc, *a,**k):
	"""\
		This code is responsible for turning whatever callable you pass in
		into a "yield from"-style coroutine.
		"""
	proc = proc(*a,**k)
	if hasattr(proc,'__await__'):
		return (yield from proc.__await__())
	if inspect.iscoroutine(proc) or inspect.isgenerator(proc):
		proc = (yield from proc)
	return proc

class RPCservice(object):
	"""\
		This object handles one specific RPC service
		"""
	queue = None
	is_alert = None
	call_conv = None
	durable = False
	ttl = None

	def __init__(self, fn,name=None, call_conv=CC_MSG, durable=None, ttl=None):
		if name is None:
			name = fn.__module__+'.'+fn.__name__
		self.fn = fn
		self.name = name
		self.call_conv = call_conv
		self.durable = bool(durable)
		self.uuid = uuidstr()
		if ttl is not None:
			self.ttl = ttl
	
	@asyncio.coroutine
	def run(self, *a,**k):
		res = (yield from coro_wrapper(self.fn,*a,**k))
		return res

	def __str__(self):
		if self.is_alert:
			n = "ALERT"
		elif self.is_alert is not None:
			n = "RPC"
		else:
			n = '?'
		return "‹%s %s %s %s›" % (n,self.name,self.call_conv,self.fn)
		



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

"""\
	This implements a bunch of mostly-generic protocol handling classes.
	"""
import asyncio
from time import time
import weakref

class Disconnected(BaseException):
	pass

class Protocol(asyncio.Protocol):
	"""\
		This class is responsible for translating the protocol's byte
		stream to messages, and vice versa.

		If you stream data out, you should periodically do
			yield from protocol.paused
		to make sure that the buffer doesn't go out of bounds.

		"""
	def __init__(self, loop=None):
		self._loop = loop if loop is not None else asyncio.get_event_loop()
		self.queue = asyncio.Queue(loop=self._loop)
		self.paused = asyncio.Future(loop=self._loop)
		self.paused.set_result(False)

	def close(self):
		self.transport.close()

	def connection_made(self, transport):
		#peername = transport.get_extra_info('peername')
		#print('Connection from {}'.format(peername))
		self.transport = transport

	def connection_lost(self, exc):
		if exc is None:
			exc = Disconnected()
		if not self.paused.done():
			self.paused.set_exception(exc) # pragma: no cover
		self.queue.put_nowait(exc)
		
	def data_received(self, data):
		try:
			for m in self.received(data):
				self.queue.put_nowait(m)
		except BaseException as exc:
			if not self.paused.done():
				self.paused.set_exception(exc)
			self.queue.put_nowait(exc)

	def received(self, data): # pragma: no cover
		"""\
			You must override this method!

			Translate the incoming byte stream to messages and yield them.
			"""
		raise NotImplementedError("You need to override %s.receive" % self.__class__.__name__)
	
	def send(self, whatever=None, *a,**k):
		"""\
			You must override this method!

			Translate the message to be sent to a bytestream
			and call self.transport.write().
			"""
		raise NotImplementedError("You need to override %s.send" % self.__class__.__name__)

	def pause_writing(self): # pragma: no cover
		self.paused = asyncio.Future(loop=self._loop)
	def resume_writing(self): # pragma: no cover
		self.paused.set_result(True)
		
class ProtocolInteraction(object):
	"""\
		A generic message read/write thing.

		You override interact() to send and receive messages.
		A client typically sends a message, waits for a reply (or more), possibly repeats, then exits.

		@asyncio.coroutine
		def interact(self):
			yield from self.paused ## periodically do this if you send lots
			self.send("Foo!")
			assert (yield from self.recv()) == "Bar?"
			
		"""

	_conn = None

	def __init__(self, *, loop=None, conn=None):
		self._protocol = None
		self._loop = loop if loop is not None else asyncio.get_event_loop()
		if conn is not None:
			self._conn = weakref.ref(conn)

	@property
	def paused(self): # pragma: no cover
		return self._paused()
	@asyncio.coroutine
	def _paused(self): # pragma: no cover
		p = self._protocol.paused
		if not p.done():
			yield from p
			self._protocol.paused.result()

	def run(self,*a, **kw):
		"""\
			If you submitted the connection while creating, you can run the interaction on this connection here.
			"""
		c = self._conn()
		if c is None:
			raise RuntimeError("Connection has gone away")
		return c.run(self,*a,**kw)
	run._is_coroutine = True

	@asyncio.coroutine
	def interact(self,*a,**k): # pragma: no cover
		raise NotImplementedError("You need to override %s.interact" % self.__class__.__name__)

	def send(self,*a,**k):
		self._protocol.send(*a,**k)

	@asyncio.coroutine
	def recv(self):
		res = (yield from self._protocol.queue.get())
		if isinstance(res,BaseException):
			raise res
		return res
		
class ProtocolClient(object):
	"""\
		A generic streaming client.

		You use this object by encapsulating a sequence of read or write
		calls in a ProtocolInteraction, then call this object's "run"
		method with it.

		This client uses multiple connections.
		"""
	MAX_IDLE = 10
	def __init__(self, protocol, host,port, loop=None):
		"""\
			@protocol: factory for the protocol to run on the connection(s)
			@host, @port: the service to talk to.
			"""
		self.protocol = protocol
		self.host = host
		self.port = port
		self.conns = []
		self._loop = loop if loop is not None else asyncio.get_event_loop()
		self._id = 1
		self.tasks = {}

	@asyncio.coroutine
	def _get_conn(self):
		now = time()
		while self.conns:
			ts,conn = self.conns.pop()
			if ts > now-self.MAX_IDLE:
				break
			assert conn.queue.empty()
			try:
				conn.close()
			except Exception: # pragma: no cover
				logger.exception("Closing idle connection")
		else:
			_,conn = (yield from self._loop.create_connection(lambda: self.protocol(loop=self._loop), self.host,self.port))
		return conn
		
	def _put_conn(self,conn):
		self.conns.append((time(),conn))

	@property
	def next_id(self):
		id = self._id
		self._id += 1
		return id

	@asyncio.coroutine
	def run(self, interaction, *a,**k):
		"""\
			Run the interaction on (an instance of) this connection.
			"""
		conn = (yield from self._get_conn())
		f = None
		id = self.next_id
		try:
			if interaction._protocol is not None:
				raise RuntimeError("%s is running twice" % repr(interaction))
			try:
				interaction._protocol = conn
				f = asyncio.ensure_future(interaction.interact(*a,**k), loop=self._loop)
				self.tasks[id] = f
				yield from f
				res = f.result()
			finally:
				assert interaction._protocol is conn
				interaction._protocol = None
		except BaseException as exc:
			if f is not None and not f.done():
				f.set_Exception(exc)
			raise
		else:
			if f is not None and not f.done():
				f.set(True) # pragma: no cover
			self._put_conn(conn)
			conn = None
			return res
		finally:
			if f is not None and not f.done():
				f.set(False) # pragma: no cover
			self.tasks.pop(id,None)
			if conn is not None:
				conn.close()

	def abort(self):
		"""Kill all tasks and connections"""
		for id,f in self.tasks.items():
			try:
				f.cancel()
			except Exception: # pragma: no cover
				pass
		while self.conns:
			_,conn = self.conns.pop()
			try:
				conn.close()
			except Exception: # pragma: no cover
				logger.exception("Trying to abort")

	@asyncio.coroutine
	def close(self):
		"""Wait for all tasks to finish"""
		while self.tasks:
			for k in list(self.tasks.keys()):
				yield from self.tasks.pop(k,None)
		self.abort()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

__VERSION__ = (0,15,2)

# Python 3.5 deprecates .async in favor of .ensure_future

# Load all these symbols from setup, because of possible load order conflicts
def setup(*a,**k):
	import qbroker.util.sync as sync
	sync.setup(*a,**k)

	import asyncio
	if not hasattr(asyncio,'ensure_future'):
		asyncio.ensure_future = asyncio.async

	global Unit,make_unit
	from qbroker.unit import Unit, make_unit

loop = None # set by setup
# unit_sync() and/or unit_gevent() will be added by qbroker.util.sync.setup()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

# Utility code

from base64 import b64encode
from collections.abc import Mapping
from importlib import import_module

import pytz
UTC = pytz.UTC
with open("/etc/localtime", 'rb') as tzfile:
	TZ = pytz.tzfile.build_tzinfo(str('local'), tzfile)

def uuidstr(u=None):
	if u is None:
		import uuid
		u=uuid.uuid1()
	return b64encode(u.bytes, altchars=b'-_').decode('ascii').rstrip('=')

def import_string(name):
	"""Import a module, or resolve an attribute of a module."""
	name = str(name)
	try:
		return import_module(name)
	except ImportError:
		if '.' not in name:
			raise
		module, obj = name.rsplit('.', 1)
		try:
			return getattr(import_string(module),obj)
		except AttributeError:
			raise AttributeError(name)

def combine_dict(*d):
	res = {}
	keys = {}
	if len(d) <= 1:
		return d
	for kv in d:
		for k,v in kv.items():
			if k not in keys:
				keys[k] = []
			keys[k].append(v)
	for k,v in keys.items():
		if len(v) == 1:
			res[k] = v[0]
		elif not isinstance(v[0],Mapping):
			for vv in v[1:]:
				assert not isinstance(vv,Mapping)
			res[k] = v[0]
		else:
			res[k] = combine_dict(*v)
	return res

class _missing: pass

class attrdict(dict):
	"""A dictionary which can be accessed via attributes, for convenience"""
	def __init__(self,*a,**k):
		super(attrdict,self).__init__(*a,**k)
		self._done = set()

	def __getattr__(self,a):
		if a.startswith('_'):
			return super(attrdict,self).__getattr__(a)
		try:
			return self[a]
		except KeyError:
			raise AttributeError(a)
	def __setattr__(self,a,b):
		if a.startswith("_"):
			super(attrdict,self).__setattr__(a,b)
		else:
			self[a]=b
	def __delattr__(self,a):
		del self[a]

# Python doesn't support setting an attribute on MethodType,
# and the thing is not subclass-able either,
# so I need to implement my own.
# Fortunately, this is reasonably easy.
from types import MethodType 
class _ClassMethodType(object):
	def __init__(self,_func,_self,_cls=None, **attrs):
		self.im_func = self.__func__ = _func
		self.im_self = self.__self__ = _self
		self.im_class = _cls
		self.__name__ = _func.__name__
		for k,v in attrs.items():
			setattr(self,k,v)
	def __call__(self,*a,**k):
		return self.__func__(self.im_self,*a,**k)

class _StaticMethodType(_ClassMethodType):
	def __call__(self,*a,**k):
		return self.__func__(*a,**k)

# Default timeout for the cache.
def format_dt(value, format='%Y-%m-%d %H:%M:%S'):
	try:
		return value.astimezone(TZ).strftime(format)
	except ValueError: ## naïve time: assume UTC
		return value.replace(tzinfo=UTC).astimezone(TZ).strftime(format)

def exported_classmethod(_fn=None,**attrs):
	"""\
		A classmethod thing which supports attributed methods.

		Usage:
			class test(object):
				@exported_classmethod(bar=42)
				def foo(cls, …):
					pass
			assert test.foo.bar == 42
		"""
	if _fn is None:
		def xfn(_fn):
			return exported_classmethod(_fn,**attrs)
		return xfn
	res = _ClassMethodAttr(_fn)
	attrs = _dab_(attrs)
	for k,v in attrs.items():
		setattr(res,k,v)
	res._attrs = attrs
	return res

def exported_staticmethod(_fn=None,**attrs):
	"""\
		A classmethod thing which supports attributed methods.

		Usage:
			class test(object):
				@exported_staticmethod(bar=42)
				def foo(…):
					pass
			assert test.foo.bar == 42
		"""
	if _fn is None:
		def xfn(_fn):
			return exported_staticmethod(_fn,**attrs)
		return xfn
	res = _StaticMethodAttr(_fn)
	attrs = _dab_(attrs)
	for k,v in attrs.items():
		setattr(res,k,v)
	res._attrs = attrs
	return res

def exported_property(fget=None,fset=None,fdel=None,doc=None,**attrs):
	"""\
		A classmethod thing which supports attributed methods.

		Usage:
			class test(object):
				@exported_staticmethod(bar=42)
				def foo(…):
					pass
			assert test.foo.bar == 42
		"""
	if fget is None and attrs is not None:
		def xfn(fget=None,fset=None,fdel=None,doc=None):
			return exported_property(fget=fget,fset=fset,fdel=fdel,doc=doc,**attrs)
		return xfn
	attrs = dict(_dab_(attrs))
	res = _PropertyAttr(fget=fget,fset=fset,fdel=fdel,doc=doc,**attrs)
	return res

class _PropertyAttr(property):
	"""Yet another rewrite in Python."""
	# The problem here is that
	#	@exported_property
	#	def x(self): …
	#	@x.setter
	#	def x(self,val): …
	# will call exported_property() TWICE. (Or thrice if you also want a deleter.)
	# Therefore we stash the attributes in the function.
	# TODO: check if one of these has attributes which are not "ours"?

	def __init__(self, fget=None,fset=None,fdel=None,doc=None,**attrs):
		property.__init__(self,fget=fget,fset=fset,fdel=fdel,doc=doc)
		f = fget
		if not attrs:
			if not f or not f.__dict__:
				f = fset
				if not f or not f.__dict__:
					f = fdel
			if f:
				attrs = f.__dict__
		for f in (fget,fset,fdel):
			if f is None:
				continue
			for k,v in attrs.items():
				setattr(f,k,v)
		for k,v in attrs.items():
			setattr(self,k,v)
		self.__doc__=doc

class _ClassMethodAttr(classmethod):
	def __get__(self,i,t=None):
		if i is not None:
			# Ignore calls on objects
			return classmethod.__get__(self,i,t)

		# Otherwise we need to build this thing ourselves.
		# TODO: use our data directly, rather than calling up.
		res = classmethod.__get__(self,i,t)
		return _ClassMethodType(res.__func__,res.__self__, **self._attrs)

class _StaticMethodAttr(classmethod):
	def __get__(self,i,t=None):
		if i is not None:
			# Ignore calls on objects
			return classmethod.__get__(self,i,t)

		# same as classmethod (almost)
		res = classmethod.__get__(self,i,t)
		return _StaticMethodType(res.__func__,res.__self__, **self._attrs)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

# Utility code

import asyncio
import inspect
import signal

import logging
logger = logging.getLogger(__name__)

iscoroutinefunction = getattr(inspect,'iscoroutinefunction', lambda _:False)

class Main:
	"""Implement a bare-bones mainloop for asyncio."""

	def __init__(self, loop=None):
		if loop is None:
			loop = asyncio.new_event_loop()
			asyncio.set_event_loop(loop)
		self.loop = loop
		self._sig = asyncio.Event(loop=loop)
		self._cleanup = []

		self.loop.add_signal_handler(signal.SIGINT,self._tilt)
		self.loop.add_signal_handler(signal.SIGTERM,self._tilt)

	@asyncio.coroutine
	def at_start(self):
		"""Called after successful startup. Overrideable."""
		yield None

	@asyncio.coroutine
	def _at_stop(self):
		"""Process cleanup code. Don't override."""
		for fn,a,k in self._cleanup[::-1]:
			try:
				if inspect.isgeneratorfunction(fn) or iscoroutinefunction(fn):
					yield from fn(*a,**k)
				else:
					fn(*a,**k)
			except Exception:
				logger.exception("Cleanup: %s %s %s",fn,repr(a),repr(k))

	def add_cleanup(self,fn,*a,**k):
		"""Register some clean-up code. Processed in reverse order."""
		self._cleanup.append((fn,a,k))
	
	def run(self):
		self.loop.run_until_complete(self._run())

	@asyncio.coroutine
	def _run(self):
		yield from self.at_start()
		yield from self._sig.wait()
		yield from self._at_stop()

	def _tilt(self):
		self.loop.remove_signal_handler(signal.SIGINT)
		self.loop.remove_signal_handler(signal.SIGTERM)
		self.loop.call_soon_threadsafe(self._tilt2)
	def _tilt2(self):
		self._sig.set()
	
	def stop(self):
		"""Stop the loop."""
		self._tilt()




#!/usr/bin/python
# -*- coding: utf-8 -*-

from __future__ import print_function,absolute_import
import sys
from time import mktime
from json.encoder import JSONEncoder
from json.decoder import JSONDecoder
from qbroker.util import attrdict, TZ,UTC, format_dt
import datetime as dt
from collections.abc import Mapping

CODEC = "application/json"

DEBUG=False

class Encoder(JSONEncoder):
	def __init__(self):
		super(Encoder,self).__init__(skipkeys=False, ensure_ascii=False,
			check_circular=True, allow_nan=False, sort_keys=False,
			indent=(2 if DEBUG else None),
			separators=((', ', ': ') if DEBUG else (',', ':')))
enc = Encoder()

def encode(data,tablespace=None):
	res = enc.encode(data)
	if isinstance(res,str):
		res = res.encode('utf-8')
	return res

class Decoder(JSONDecoder):
	pass
dec = Decoder()

def decode(data, tablespace=None, proxy=None, p1=None):
	if isinstance(data,bytes):
		data = data.decode('utf-8')
	return dec.decode(data)




#!/usr/bin/python
# -*- coding: utf-8 -*-

from __future__ import print_function,absolute_import
import sys
from time import mktime
from json.encoder import JSONEncoder
from json.decoder import JSONDecoder
from qbroker.util import attrdict, TZ,UTC, format_dt
import base64
import datetime as dt
from collections.abc import Mapping

CODEC = "application/json+obj"
DEFAULT = True

DEBUG=False

type2cls = {}
name2cls = {}
def _reg(cls):
	type2cls[cls.cls.__name__] = cls
	name2cls[cls.clsname] = cls
	return cls

@_reg
class _binary(object):
	"""A codec for bytes.
		Try to convert to an utf-8 string; if not possible, use base64.a85."""
	cls = bytes
	clsname = "bin"

	@staticmethod
	def encode(obj):
		## the string is purely for human consumption and therefore does not have a time zone
		try:
			obj = obj.decode('utf-8',errors='strict')
		except Exception:
			return {"b":base64.a85encode(obj).decode('utf-8'), "s":obj.decode('utf-8',errors='ignore')}
		else:
			return {"s":obj}

	@staticmethod
	def decode(b=None,s=None):
		if b is None:
			return s.encode('utf-8')
		else:
			return base64.a85decode(b)

@_reg
class _datetime(object):
	cls = dt.datetime
	clsname = "datetime"

	@staticmethod
	def encode(obj):
		## the string is purely for human consumption and therefore does not have a time zone
		return {"t":mktime(obj.timetuple()),"s":format_dt(obj)}

	@staticmethod
	def decode(t=None,s=None,a=None,k=None,**_):
		if t:
			return dt.datetime.utcfromtimestamp(t).replace(tzinfo=UTC).astimezone(TZ)
		else: ## historic
			assert a
			return dt.datetime(*a).replace(tzinfo=TZ)

@_reg
class _date(object):
	cls = dt.date
	clsname = "date"

	@staticmethod
	def encode(obj):
		return {"d":obj.toordinal(), "s":obj.strftime("%Y-%m-%d")}

	@staticmethod
	def decode(d=None,s=None,a=None,**_):
		if d:
			return dt.date.fromordinal(d)
		## historic
		return dt.date(*a)

@_reg
class _time(object):
	cls = dt.time
	clsname = "time"

	@staticmethod
	def encode(obj):
		ou = obj.replace(tzinfo=UTC)
		secs = ou.hour*3600+ou.minute*60+ou.second
		return {"t":secs,"s":"%02d:%02d:%02d" % (ou.hour,ou.minute,ou.second)}

	@staticmethod
	def decode(t=None,s=None,a=None,k=None,**_):
		if t:
			return dt.datetime.utcfromtimestamp(t).time()
		return dt.time(*a)

class RemoteException(Exception):
	def _init__(self,r,s):
		self.r = r
		self.s = s
	def __str__(self):
		return self.s
	def __repr__(self):
		return self.r

@_reg
class _exc(object):
	cls = BaseException
	clsname = "exc"

	@staticmethod
	def encode(obj):
		if isinstance(obj,Exception):
			return {"e":obj.__class__.__name__,"s":str(obj),"r":repr(obj)}
		else:
			return {"b":obj.__class__.__name__,"s":str(obj),"r":repr(obj)}

	@staticmethod
	def decode(b=None,e=None,r=None,s=None,**_):
		if b is not None:
			return getattr(__builtins__,b)()
		return RemoteException(r,s)

class Encoder(JSONEncoder):
	def __init__(self,main=()):
		self.objcache = {}
		self.main = main
		super(Encoder,self).__init__(skipkeys=False, ensure_ascii=False,
			check_circular=False, allow_nan=False, sort_keys=False,
			indent=(2 if DEBUG else None),
			separators=((', ', ': ') if DEBUG else (',', ':')))

	def default(self, data):
		if isinstance(data,BaseException):
			obj = _exc
		else:
			obj = type2cls.get(data.__class__.__name__,None)
		if obj is not None:
			data = obj.encode(data)
			data["_o"] = obj.clsname
			return data

		if hasattr(data,"_read"):
			ci = (data._t.name,data._id)
			cr = self.objcache.get(ci,None)
			if cr is not None:
				d = {"_cr":cr}
			else:
				cr = len(self.objcache)+1
				self.objcache[ci] = cr

				d = {"_index": data._id, "_table": data._t.name, "_cr": cr}
				send = False
				if data in self.main:
					send = True
				else:
					mode = getattr(data,'_mode',F_STANDARD)
					if mode in (F_SQL,F_STORE):
						send = True
				if isinstance(data._d,Exception):
					d["_error"] = str(data._d)
				elif send:
					for k,v in data:
						d[k] = v
			return d
		return super(Encoder,self).default(data)

def encode(data):
	main = set()
	if hasattr(data,"_read"):
		if data._d is None:
			data._read()
		main.add(data)
	else:
		try:
			d = data["data"]
			if d._d is None:
				d._read()
			if isinstance(d,(list,tuple)):
				for dd in d:
					main.add(dd)
			else:
				main.add(d)
		except (TypeError,KeyError,AttributeError):
			pass
	res = Encoder(main).encode(data)
	if isinstance(res,str):
		res = res.encode('utf-8')
	return res

class Decoder(JSONDecoder):
	def __init__(self, proxy):
		self.objcache = {}
		self.proxy = proxy
		super(Decoder,self).__init__(object_hook=self.hook)

	def hook(self,data):
		if not isinstance(data,Mapping):
			return data

		ev = data.pop('_o',None)
		if ev is not None:
			return name2cls[ev].decode(**data)

		cr = data.pop('_cr',None)
		if cr is None:
			data = attrdict(data)
			return data
		d = self.objcache.get(cr,None)
		if d is None:
			d = self.proxy()
			self.objcache[cr]=d

		try:
			t = data.pop('_table')
			idx = data.pop('_index')
			idx = tuple(tuple(x) for x in idx) # convert from nested list
		except KeyError:
			pass # recursive or whatever
		else:
			d = d._set(t,idx,**data)
			self.objcache[cr] = d
		return d
	
def decode(data, proxy=None, p1=None):
	if isinstance(data,bytes):
		data = data.decode('utf-8')

	d = Decoder(proxy)
	if p1:
		d.objcache[1] = p1
	return d.decode(data)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

import weakref
import asyncio
import aioamqp
import functools
import json
from collections.abc import Mapping

from .msg import _RequestMsg,PollMsg,RequestMsg,BaseMsg
from .rpc import CC_DICT,CC_DATA
from ..codec import get_codec
from ..util import import_string

import logging
logger = logging.getLogger(__name__)

class DeadLettered(RuntimeError):
	pass

class _ch(object):
	"""Helper object"""
	channel = None
	exchange = None
	queue = None

class Connection(object):
	amqp = None # connection
	alert_bc = False

	def __init__(self,unit,codec=None):
		if codec is None:
			codec = unit.config['amqp'].get('codec', 'DEFAULT')
		if isinstance(codec,str):
			codec = get_codec(codec)
		self._loop = unit._loop
		self.rpcs = {}
		self.alerts = {}
		self.replies = {}
		self.unit = weakref.ref(unit)
		cfg = unit.config['amqp']['server']
		if 'connect_timeout' in cfg:
			cfg['connect_timeout'] = float(cfg['connect_timeout'])
		if 'ssl' in cfg and isinstance(cfg['ssl'],str):
			cfg['ssl'] = cfg['ssl'].lower() == 'true'
		if 'port' in cfg:
			cfg['port'] = int(cfg['port'])
		self.cfg = cfg
		self.codec = codec

	@asyncio.coroutine
	def connect(self):
		logger.debug("Connecting %s",self)
		try:
			self.amqp_transport,self.amqp = (yield from aioamqp.connect(loop=self._loop, protocol_factory=NotifyingAmqpProtocol, **self.cfg))
			#self.amqp_transport,self.amqp = (yield from aioamqp.connect(loop=self._loop, **self.cfg))
		except Exception as e:
			logger.exception("Not connected to AMPQ: host=%s vhost=%s user=%s", self.cfg['host'],self.cfg['virtualhost'],self.cfg['login'])
			raise
		self.amqp._init_futures(self._loop)
		yield from self.setup_channels()
		logger.debug("Connected %s",self)

	@asyncio.coroutine
	def _setup_one(self,name,typ,callback=None, q=None, route_key=None, exclusive=None, alt=None):
		"""\
			Register a channel. Internal helper.
			"""
		unit = self.unit()
		cfg = unit.config['amqp']
		ch = _ch()
		setattr(self,name,ch)
		logger.debug("setup RPC for %s",name)
		ch.channel = (yield from self.amqp.channel())
		ch.exchange = cfg['exchanges'][name]
		logger.debug("Chan %s: exchange %s", ch.channel,cfg['exchanges'][name])
		if exclusive is None:
			exclusive = (q is not None)
		d = {}
		if alt is not None:
			d["alternate-exchange"] = cfg['exchanges'][alt]
		yield from ch.channel.exchange_declare(cfg['exchanges'][name], typ, auto_delete=False, passive=False, arguments=d)

		if q is not None:
			assert callback is not None
			d = {}
			ttl = cfg['ttl'].get(name,0)
			if ttl:
				d["x-dead-letter-exchange"] = cfg['queues']['dead']+q
				d["x-message-ttl"] = cfg['ttl'][name]
			ch.queue = (yield from ch.channel.queue_declare(cfg['queues'][name]+q, auto_delete=True, passive=False, exclusive=exclusive, arguments=d))
			yield from ch.channel.basic_qos(prefetch_count=1,prefetch_size=0,connection_global=False)
			logger.debug("Chan %s: read %s", ch.channel,cfg['queues'][name]+q)
			yield from ch.channel.basic_consume(queue_name=cfg['queues'][name]+q, callback=callback)
			if route_key is not None:
				logger.debug("Chan %s: bind %s %s %s", ch.channel,cfg['exchanges'][name], route_key, ch.queue['queue'])
				yield from ch.channel.queue_bind(ch.queue['queue'], cfg['exchanges'][name], routing_key=route_key)
				pass
		else:
			assert callback is None

		logger.debug("setup RPC for %s done",name)

	@asyncio.coroutine
	def setup_channels(self):
		"""Configure global channels"""
		u = self.unit()
		# See doc/qbroker.rst
		yield from self._setup_one("alert",'topic', self._on_alert, u.uuid)
		yield from self._setup_one("rpc",'topic', self._on_rpc, u.uuid, 'qbroker.uuid.'+u.uuid, alt="dead")
		yield from self._setup_one("reply",'direct', self._on_reply, u.uuid, u.uuid)
		if u.config['amqp']['handlers']['dead']:
			yield from self._setup_one("dead",'topic', self._on_dead_rpc, "rpc", exclusive=False, route_key='#')

	@asyncio.coroutine
	def _on_dead_rpc(self, channel,body,envelope,properties):
		try:
			msg = get_codec(properties.content_type).decode(body)
			msg = BaseMsg.load(msg,envelope,properties)
			reply = msg.make_response()
			reply_to = getattr(msg, 'reply_to',None)
			reply.set_error(DeadLettered(envelope.exchange_name), envelope.routing_key, "reply")
			reply,props = reply.dump(self)
			reply = self.codec.encode(reply)
			yield from self.reply.channel.publish(reply, self.reply.exchange, reply_to, properties=props)
		finally:
			yield from channel.basic_client_ack(envelope.delivery_tag)
		
	@asyncio.coroutine
	def _on_alert(self, channel,body,envelope,properties):
		logger.debug("read alert message %s",envelope.delivery_tag)
		try:
			msg = get_codec(properties.content_type).decode(body)
			msg = BaseMsg.load(msg,envelope,properties)
			try:
				rpc = self.alerts[msg.routing_key]
			except KeyError:
				n = msg.routing_key
				while True:
					i = n.rfind('.')
					if i < 1:
						rpc = self.alerts.get('#',None)
						if rpc is not None:
							break
						raise
					n = n[:i]
					rpc = self.alerts.get(n+'.#',None)
					if rpc is not None:
						break
			if rpc.call_conv == CC_DICT:
				a=(); k=msg.data
				if not isinstance(k,Mapping):
					assert k == ''
					k = {}
			elif rpc.call_conv == CC_DATA:
				a=(msg.data,); k={}
			else:
				a=(msg,); k={}

			reply_to = getattr(msg, 'reply_to',None)
			if reply_to:
				try:
					data = (yield from rpc.run(*a,**k))
				except Exception as exc:
					reply = msg.make_response()
					logger.exception("error on alert %s: %s", envelope.delivery_tag, body)
					reply.set_error(exc, rpc.name,"reply")
				else:
					if data is None:
						return
					reply = msg.make_response()
					reply.data = data
				reply,props = reply.dump(self)
				reply = self.codec.encode(reply)
				yield from self.reply.channel.publish(reply, self.reply.exchange, reply_to, properties=props)
			else:
				try:
					yield from rpc.run(*a,**k)
				except Exception as exc:
					logger.exception("error on alert %s: %s", envelope.delivery_tag, body)

		except Exception as exc:
			logger.exception("problem with rpc %s: %s", envelope.delivery_tag, body)
			yield from self.alert.channel.basic_reject(envelope.delivery_tag)
		else:
			logger.debug("ack rpc %s",envelope.delivery_tag)
			yield from channel.basic_client_ack(envelope.delivery_tag)

	@asyncio.coroutine
	def _on_rpc(self, channel,body,envelope,properties):
		logger.debug("read rpc message %s",envelope.delivery_tag)
		try:
			msg = get_codec(properties.content_type).decode(body)
			msg = BaseMsg.load(msg,envelope,properties)
			rpc = self.rpcs[msg.routing_key]
			reply = msg.make_response()
			try:
				if rpc.call_conv == CC_DICT:
					a=(); k=msg.data
				elif rpc.call_conv == CC_DATA:
					a=(msg.data,); k={}
				else:
					a=(msg,); k={}
				reply.data = (yield from rpc.run(*a,**k))
			except Exception as exc:
				logger.exception("error on rpc %s: %s", envelope.delivery_tag, body)
				reply.set_error(exc, rpc.name,"reply")
			reply,props = reply.dump(self)
			reply = self.codec.encode(reply)
			yield from channel.publish(reply, self.reply.exchange, msg.reply_to, properties=props)
		except Exception as exc:
			logger.exception("problem with rpc %s: %s", envelope.delivery_tag, body)
			yield from channel.basic_reject(envelope.delivery_tag)
		else:
			logger.debug("ack rpc %s",envelope.delivery_tag)
			yield from channel.basic_client_ack(envelope.delivery_tag)

	@asyncio.coroutine
	def _on_reply(self, channel,body,envelope,properties):
		logger.debug("read reply message %s",envelope.delivery_tag)
		try:
			msg = get_codec(properties.content_type).decode(body)
			msg = BaseMsg.load(msg,envelope,properties)
			f,req = self.replies[msg.correlation_id]
			try:
				yield from req.recv_reply(f,msg)
			except Exception as exc: # pragma: no cover
				if not f.done():
					f.set_exception(exc)
		except Exception as exc:
			yield from self.reply.channel.basic_reject(envelope.delivery_tag)
			logger.exception("problem with message %s: %s", envelope.delivery_tag, body)
		else:
			logger.debug("ack message %s",envelope.delivery_tag)
			yield from channel.basic_client_ack(envelope.delivery_tag)

	@asyncio.coroutine
	def call(self,msg, timeout=None, dest=None):
		if dest is None:
			dest = msg.routing_key
		cfg = self.unit().config['amqp']
		if timeout is None:
			tn = getattr(msg,'_timer',None)
			if tn is not None:
				timeout = self.unit().config['amqp']['timeout'].get(tn,None)
				if timeout is not None:
					timeout = float(timeout)
		assert isinstance(msg,_RequestMsg)
		data,props = msg.dump(self)
		data = self.codec.encode(data)
		if timeout is not None:
			f = asyncio.Future(loop=self._loop)
			id = msg.message_id
			self.replies[id] = (f,msg)
		logger.debug("Send %s to %s: %s", dest, cfg['exchanges'][msg._exchange], data)
		yield from getattr(self,msg._exchange).channel.publish(data, cfg['exchanges'][msg._exchange], dest, properties=props)
		if timeout is None:
			return
		try:
			yield from asyncio.wait_for(f,timeout, loop=self._loop)
		except asyncio.TimeoutError:
			if isinstance(msg,PollMsg):
				return msg.replies
			raise # pragma: no cover
		finally:
			del self.replies[id]
		return f.result()
		
	@asyncio.coroutine
	def register_rpc(self,rpc):
		ch = self.rpc
		cfg = self.unit().config['amqp']
		assert rpc.name not in self.rpcs
		assert rpc.queue is None
		rpc.channel = (yield from self.amqp.channel())
		d = {}
		if cfg['ttl']['rpc'] or rpc.ttl:
			d["x-dead-letter-exchange"] = cfg['queues']['dead']+'rpc'
			d["x-message-ttl"] = rpc.ttl if rpc.ttl else cfg['ttl']['rpc']
		rpc.queue = (yield from rpc.channel.queue_declare(cfg['queues']['rpc']+rpc.name.replace('.','_'), auto_delete=not rpc.durable, passive=False, arguments=d))
		logger.debug("Chan %s: bind %s %s %s", ch.channel,cfg['exchanges']['rpc'], rpc.name, rpc.queue['queue'])
		yield from rpc.channel.queue_bind(rpc.queue['queue'], cfg['exchanges']['rpc'], routing_key=rpc.name)
		self.rpcs[rpc.name] = rpc

		yield from rpc.channel.basic_qos(prefetch_count=1,prefetch_size=0,connection_global=False)
		logger.debug("Chan %s: read %s", rpc.channel,rpc.queue['queue'])
		yield from rpc.channel.basic_consume(queue_name=rpc.queue['queue'], callback=self._on_rpc, consumer_tag=rpc.uuid)

	@asyncio.coroutine
	def unregister_rpc(self,rpc):
		ch = self.rpc
		cfg = self.unit().config['amqp']
		if isinstance(rpc,str):
			rpc = self.rpcs.pop(rpc)
		else:
			del self.rpcs[rpc.name]
		assert rpc.queue is not None
		logger.debug("Chan %s: unbind %s %s %s", ch.channel,cfg['exchanges']['rpc'], rpc.name, rpc.queue['queue'])
		yield from rpc.channel.queue_unbind(rpc.queue['queue'], cfg['exchanges']['rpc'], routing_key=rpc.name)
		logger.debug("Chan %s: noread %s", rpc.channel,rpc.queue['queue'])
		yield from rpc.channel.basic_cancel(consumer_tag=rpc.uuid)

	@asyncio.coroutine
	def register_alert(self,rpc):
		assert rpc.name not in self.alerts
		n = rpc.name
		if rpc.name.endswith('.#'):
			n = n[:-2]
		if len(n) > 1 and '#' in n:
			raise RuntimeError("I won't find that")

		ch = self.alert
		cfg = self.unit().config['amqp']
		logger.debug("Chan %s: bind %s %s %s", ch.channel,cfg['exchanges']['alert'], rpc.name, ch.exchange)
		yield from ch.channel.queue_bind(ch.queue['queue'], ch.exchange, routing_key=rpc.name)
		self.alerts[rpc.name] = rpc

		if rpc.name.endswith('.#'):
			self.alert_bc = True

	@asyncio.coroutine
	def unregister_alert(self,rpc):
		if isinstance(rpc,str):
			rpc = self.alerts.pop(rpc)
		else:
			del self.alerts[rpc.name]
		ch = self.alert
		cfg = self.unit().config['amqp']
		logger.debug("Chan %s: unbind %s %s %s", ch.channel,cfg['exchanges']['alert'], rpc.name, ch.exchange)
		yield from ch.channel.queue_unbind(ch.queue['queue'], ch.exchange, routing_key=rpc.name)

	@asyncio.coroutine
	def close(self):
		a,self.amqp = self.amqp,None
		if a is not None:
			logger.debug("Disconnecting %s",self)
			try:
				yield from a.close(timeout=1)
			except Exception: # pragma: no cover
				logger.exception("closing the connection")
			self.amqp_transport = None
			logger.debug("Disconnected %s",self)

	def _kill(self):
		self.amqp = None
		a,self.amqp_transport = self.amqp_transport,None
		if a is not None:
			logger.debug("Killing %s",self)
			try:
				a.close()
			except Exception: # pragma: no cover
				logger.exception("killing the connection")

import aioamqp.protocol
class NotifyingAmqpProtocol(aioamqp.protocol.AmqpProtocol):
	"""Adds a future that triggers when the protocol gets disconnected"""
	when_disconnected = None

	def _init_futures(self,loop):
		self.when_disconnected = asyncio.Future(loop=loop)

	def connection_made(self, exc):
		if self.when_disconnected is not None and self.when_disconnected.done():
			self.when_disconnected = asyncio.Future(loop=self.when_disconnected._loop)
		super().connection_made(exc)

	def connection_lost(self, exc):
		super().connection_lost(exc)
		if self.when_disconnected is not None:
			if exc is None:
				self.when_disconnected.set_result(None)
			else:
				self.when_disconnected.set_exception(exc)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

"""\
This module implements accessors for using QBroker in a threaded environment,
either natively or via gevent.

For native threads, call `qbroker.setup(sync=True)` and use `*_sync` methods.
You can run asyncio code thus:
>>> from qbroker.util.sync import AioRunner
>>> result = await_sync(proc,*a,**kw)

For gevent threads, call `qbroker.setup(gevent=True)` and use `*_gevent` methods.
You can run asyncio code thus:
>>> from qbroker.util.sync import await_from
>>> result = await_gevent(proc,*a,**kw)

Note that, unlike these functions, native asyncio calls the procedure directly:
>>> res = (yield from proc(*a,**k)) ## Python 3.4
>>> res = await proc(*a,**k)        ## Python 3.5+
"""
# Utility code

import os
import qbroker
import sys

# overwritten by setup(), if applicable
aiogevent = None
threading = None
asyncio = None

class AioRunner:
	"""A singleton which supplies a thread for running asyncio tasks.

	Call AioRunner.start(setup,teardown) to set things up; these must
	be argument-less coroutines which are executed in the new task.
	If the loop is already running, @setup is scheduled immediately.

	Call AioRunner.stop() to halt things. All registered teardown functions
	will be called in reverse order.

	Calls to .start() and .stop() must be balanced.

	Exceptions in setup will be propagated.
	Exceptions in teardown are logged but otherwise ignored.

	You need to call qbroker.setup(sync=True) before this is useable.

	NOTE: this is not useful for gevent threads!
	Instead, call qbroker.setup(gevent=True) and use the `aiogevent` module.
	"""

	def __init__(self):
		self.lock = threading.Lock()
		self._cleanup()

	def start(self, setup=None,teardown=None):
		with self.lock:
			if self._loop is not None:
				self.run_async(setup)
				if teardown is not None:
					self.teardown.append(teardown)
				self.started += 1
				return

			self.setup = setup

			from concurrent import futures
			self.done = futures.Future()
			self.ready = futures.Future()
			self.thread = threading.Thread(target=self._runner)
			self.thread.start()
			try:
				res = self.ready.result()
			except Exception as ex:
				self.thread.join()
				self._cleanup()
				raise
			else:
				if teardown is not None:
					self.teardown.append(teardown)
				return res
			finally:
				del self.ready

	def stop(self):
		with self.lock:
			assert self._loop is not None, "not started"
			if self.started > 0:
				self.started -= 1
				return
			self._loop.call_soon_threadsafe(self.end.set_result,None)
			try:
				self.done.result()
			finally:
				if self._loop is not None:
					self._loop.close()
				if self.thread is not None:
					self.thread.join()
				self._cleanup()

	def _cleanup(self):
		self._loop = None
		self.thread = None
		self.end = None
		self.done = None
		self.started = 0
		self.setup = None
		self.teardown =[]

	@property
	def loop(self):
		if self._loop is None:
			self.start()
		return self._loop

	def _runner(self):
		self._loop = asyncio.new_event_loop()
		asyncio.set_event_loop(self._loop)
		self.end = asyncio.Future(loop=self._loop)

		@asyncio.coroutine
		def _worker():
			try:
				if self.setup is not None:
					yield from self.setup()
			except Exception as ex:
				self.ready.set_exception(ex)
			else:
				self.ready.set_result(None)
				yield from self.end
			for fn in self.teardown[::-1]:
				try:
					yield from fn()
				except Exception as ex:
					logger.exception("Error during %s", fn)

		try:
			self._loop.run_until_complete(_worker())
		except Exception as ex:
			self.done.set_exception(ex)
		else:
			self.done.set_result(None)
			
	def run_async(self,proc,*args, _async=False,_timeout=None, **kwargs):
		"""Run an asyncio-using procedure.
		
		@_async: set if you want to run it in the background and get the
				future returned instead.
		@_timeout: set if you want the procedure cancelled if it takes
				longer than that many seconds.

		All other parameters will be passed to @proc.
		
		"""
		from concurrent import futures

		def runner(fx):
			try:
				f = asyncio.ensure_future(proc(*args,**kwargs), loop=self.loop)
			except BaseException as exc:
				fx.set_exception(exc)
			else:
				if _timeout is not None:
					f = asyncio.ensure_future(asyncio.wait_for(f,_timeout,loop=self.loop), loop=self.loop)
				def done(ff):
					assert f is ff, (f,ff)
					try:
						fx.set_result(f.result())
					except Exception as exc:
						fx.set_exception(exc)
				f.add_done_callback(done)

		fx = futures.Future()
		AioRunner.loop.call_soon_threadsafe(runner,fx)
		if _async:
			return fx
		else:
			return fx.result()

def setup(sync=False,gevent=False):
	if gevent and not qbroker.loop:
		## You get spurious errors if the core threading module is imported
		## before monkeypatching.
		#	if 'threading' in sys.modules and 'TRAVIS' not in os.environ:
		#		raise Exception('The ‘threading’ module was loaded before patching for gevent')
		## However, simply not doing it at all is more prudent:
		## Qbroker itself no longer needs it.
		#	import gevent.monkey
		#	gevent.monkey.patch_all()
		pass

	global asyncio
	import asyncio

	if gevent and not qbroker.loop:
		global aiogevent
		import aiogevent
		asyncio.set_event_loop_policy(aiogevent.EventLoopPolicy())
		qbroker.loop = asyncio.get_event_loop()

		def make_unit_gevent(*args,**kwargs):
			return aiogevent.yield_future(asyncio.ensure_future(qbroker.make_unit(*args, loop=qbroker.loop, **kwargs), loop=qbroker.loop))
		qbroker.make_unit_gevent = make_unit_gevent

	global AioRunner
	if sync and isinstance(AioRunner,type):
		# The class is at top level when not in use, so you can use pydoc on it.
		# Here we replace the class with a singleton instance.
		global threading
		import threading
		AioRunner = AioRunner()

		def make_unit_sync(*a,**k):
			return AioRunner.run_async(qbroker.make_unit, *args,**kwargs)
		qbroker.make_unit_sync = make_unit_sync
	
class SyncFuncs(type):
	""" A metaclass which adds synchronous version of coroutines.

	This metaclass finds all coroutine functions defined on a class
	and adds a synchronous version with a '_sync' suffix appended to the
	original function name.

	The sync version will behave as if it were called via
	`AioRunner.run_async`, including its _async and _timeout arguments.

	The gevent version will behave as if it were called via
	`aiogevent.yield_future`, including its _timeout arguments.

	"""
	def __new__(cls, clsname, bases, dct, **kwargs):
		new_dct = {}

		global asyncio
		if asyncio is None:
			import asyncio

		for name,val in dct.items():
			# Make a sync version of all coroutine functions
			if asyncio.iscoroutinefunction(val):
				meth = sync_maker(name)
				syncname = '{}_sync'.format(name)
				meth.__name__ = syncname
				meth.__qualname__ = '{}.{}'.format(clsname, syncname)
				new_dct[syncname] = meth

				meth = gevent_maker(name)
				syncname = '{}_gevent'.format(name)
				meth.__name__ = syncname
				meth.__qualname__ = '{}.{}'.format(clsname, syncname)
				new_dct[syncname] = meth
		dct.update(new_dct)

		return super().__new__(cls, clsname, bases, dct)

def sync_maker(func):
	def sync_func(self, *args, **kwargs):
		meth = getattr(self, func)
		return AioRunner.run_async(meth, *args,**kwargs)
	return sync_func

def gevent_maker(func):
	def gevent_func(self, *args, _timeout=None, _async=False, **kwargs):
		meth = getattr(self, func)
		f = asyncio.ensure_future(meth(*args,**kwargs), loop=qbroker.loop)
		if _timeout is not None:
			f = asyncio.ensure_future(asyncio.wait_for(f,_timeout,loop=qbroker.loop), loop=qbroker.loop)
		if _async:
			return gevent.spawn(aiogevent.yield_future,f)
		else:
			return aiogevent.yield_future(f)
	return gevent_func

def await_sync(proc,*a,**k):
	"""Helper function to wait for an async result from a native thread"""
	return AioRunner.run_async(proc,*a,**k)

def await_gevent(proc,*a, _loop=None, **k):
	"""Helper function to wait for an async result from a gevent thread"""
	if _loop is None:
		_loop = qbroker.loop
	p = proc(*a,**k)
	f = asyncio.ensure_future(p, loop=_loop)
	return aiogevent.yield_future(f, loop=_loop)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

"""\
	This code implements a simple line-based protocol.
	"""

import asyncio
import struct
from time import time
from . import Protocol

import logging
logger = logging.getLogger(__name__)

class LineProtocol(Protocol):
	MAX_LENGTH=10*1024
	LF=b'\n'

	def __init__(self, loop=None):
		super().__init__(loop=loop)
		self.data = b""

	def received(self, data):
		self.data += data
		while True:
			lf = self.data.find(self.LF)
			if lf == -1:
				break
			data = self.data[:lf]
			self.data = self.data[lf+len(self.LF):]
			yield data.decode('utf-8')
		if len(self.data) > self.MAX_LENGTH:
			raise RuntimeError("message too long in %s" % repr(self()))

	def send(self, data):
		self.transport.write(data.encode('utf-8') + self.LF)




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

import asyncio
from ..util import uuidstr, combine_dict
from ..util.sync import SyncFuncs, sync_maker,gevent_maker
from .msg import RequestMsg,PollMsg,AlertMsg
from .rpc import CC_MSG,CC_DATA
from . import DEFAULT_CONFIG
from collections.abc import Mapping
from aioamqp.exceptions import ChannelClosed

import logging
logger = logging.getLogger(__name__)

class _NOTGIVEN:
	pass

class Unit(object, metaclass=SyncFuncs):
	"""The basic QBroker messenger."""
	config = None # configuration data
	conn = None # AMQP receiver
	uuid = None # my UUID
	restarting = None
	args = ()

	def __init__(self, app, *, loop=None, **cfg):
		"""\
			Connect to an AMQP server. See qbroker.unit.DEFAULT_CONFIG for
			all recognized parameters.

			>>> u = Unit("my_nice_server", 
			...      amqp=dict(
			...         server=dict(
			...            login="foo",
			...            password="bar",
			...            virtualhost="/test")))

			You need to call .start() to actually initiate a connection.
			"""

		self._loop = loop or asyncio.get_event_loop()
		self.app = app

		self.config = combine_dict(cfg, DEFAULT_CONFIG)
		self.restarting = asyncio.Event(loop=loop)

		self.rpc_endpoints = {}
		self.alert_endpoints = {}

	@asyncio.coroutine
	def start(self, *args, restart=False):
		"""Connect. This may fail."""

		if self.uuid is None:
			self.uuid = uuidstr()
		if args:
			self.args = args

		self.register_alert("qbroker.ping",self._alert_ping, call_conv=CC_DATA)
		self.register_rpc("qbroker.ping", self._reply_ping)
		self.register_alert("qbroker.uuid."+self.uuid, self._alert_ping, call_conv=CC_DATA)
		self.register_rpc("qbroker.uuid."+self.uuid, self._reply_ping)
		self.register_alert("qbroker.app."+self.app, self._alert_ping, call_conv=CC_DATA)
		self.register_rpc("qbroker.app."+self.app, self._reply_ping)

		yield from self._create_conn()
		yield from self.alert('qbroker.restart' if restart else 'qbroker.start', uuid=self.uuid, app=self.app, args=args)
		self.restarting.set()
	
	@asyncio.coroutine
	def restart(self, t_min=10,t_inc=20,t_max=100):
		"""Reconnect. This will not fail."""
		while self.restarting is not None:
			yield from self.restarting.wait()

		try:
			self.close()
		except Exception:
			pass
		
		while True:
			try:
				yield from self.start(restart=True)
			except Exception as exc:
				logger.exception("Reconnect of %s failed", self)
				yield from asyncio.sleep(t_min, loop=self._loop)
				t_min = min(t_min+t_inc, t_max)
			else:
				return
	
	@asyncio.coroutine
	def stop(self, rc=0):
		if self.conn is None:
			return

		try:
			yield from self.alert('qbroker.stop', uuid=self.uuid, exitcode=rc)
		except ChannelClosed:
			pass

		self.close()
	
	## client

	@asyncio.coroutine
	def rpc(self,name, _data=None, *, _dest=None,_uuid=None, **data):
		"""Send a RPC request.
		Returns the response. 
		"""
		if _data is not None:
			assert not data, data
			data = _data
		msg = RequestMsg(name, self, data)
		assert _uuid is None or _dest is None
		if _dest is not None:
			_dest = 'qbroker.app.'+_dest
		elif _uuid is not None:
			_dest = 'qbroker.uuid.'+_uuid
		res = (yield from self.conn.call(msg,dest=_dest))
		res.raise_if_error()
		return res.data

	@asyncio.coroutine
	def alert(self,name, _data=None, *, _dest=None,_uuid=None, timeout=None,callback=None,call_conv=CC_MSG, **data):
		"""Send a broadcast alert.
		If @callback is not None, call on each response until the time runs out
		"""
		if _data is not None:
			assert not data
			data = _data
		if callback:
			msg = PollMsg(name, self, data=data, callback=callback,call_conv=call_conv)
		else:
			msg = AlertMsg(name, self, data=data)
		assert _uuid is None or _dest is None
		if _dest is not None:
			_dest = 'qbroker.uuid.'+_dest
		elif _uuid is not None:
			_dest = 'qbroker.uuid.'+_uuid
		res = (yield from self.conn.call(msg, dest=_dest, timeout=timeout))
		return res
		
	## server

	def register_rpc(self, *a, _async=False, _alert=False, call_conv=CC_MSG, durable=None,ttl=None):
		"""\
			Register an RPC listener.
				
				conn.register_rpc(RPCservice(fn,name))
				conn.register_rpc(name,fn)
				conn.register_rpc(fn)
				@conn_register(name)
				def fn(…): pass
				@conn_register
				def fn(…): pass

			If connected, you must use the _async version.
			"""
		name = None
		@asyncio.coroutine
		def reg_async(fn,epl):
			if _alert:
				yield from self.conn.register_alert(fn)
			else:
				yield from self.conn.register_rpc(fn)
			epl[name] = fn
			return fn
		def reg(fn):
			nonlocal name
			from .rpc import RPCservice
			if not isinstance(fn,RPCservice):
				if name is None:
					name = fn.__name__
					name = name.replace('._','.')
					name = name.replace('_','.')
				fn = RPCservice(name=name,fn=fn, call_conv=call_conv, durable=durable,ttl=ttl)
			else:
				if name is None:
					name = fn.name
				if durable is not None:
					assert durable is fn.durable
				if ttl is not None:
					assert ttl == fn.ttl
			assert fn.is_alert is None
			if _alert:
				epl = self.alert_endpoints
			else:
				epl = self.rpc_endpoints
			assert name not in epl, name
			fn.is_alert = _alert
			if _async and self.conn is not None:
				return reg_async(fn,epl)
			else:
				assert self.conn is None,"Use register_*_async() when connected"
			epl[name] = fn
			return fn.fn
		assert len(a) <= 2
		if len(a) == 0:
			return reg
		elif len(a) == 2:
			name = a[0]
			assert isinstance(name,str), a
			return reg(a[1])
		else:
			a = a[0]
			if isinstance(a,str):
				name = a
				return reg
			else:
				return reg(a)
	
	def register_rpc_async(self, *a, **kw):
		"""Register an RPC listener while connected.
				
				yield from conn.register_rpc(RPCservice(fn,name))
				yield from conn.register_rpc(name,fn)
				yield from conn.register_rpc(fn)
			"""
		return self.register_rpc(*a, _async=True, **kw)
	register_rpc_sync = sync_maker("register_rpc_async")
	register_rpc_gevent = gevent_maker("register_rpc_async")

	def register_alert(self, *a, _async=False, **kw):
		"""Register an alert listener.
		   See register_rpc for calling conventions."""
		return self.register_rpc(*a, _async=_async, _alert=True, **kw)

	def register_alert_async(self, *a, **kw):
		"""Register an alert listener.
		   See register_rpc_async for calling conventions."""
		return self.register_rpc(*a, _async=True, _alert=True, **kw)
	register_alert_sync = sync_maker("register_alert_async")
	register_alert_gevent = gevent_maker("register_alert_async")

	def unregister_rpc(self, fn, _async=False,_alert=False):
		if not isinstance(fn,str):
			if hasattr(fn,'name'):
				fn = fn.name
			else:
				fn = fn.__name__
				fn = fn.replace('._','.')
				fn = fn.replace('_','.')
		if _alert:
			epl = self.alert_endpoints
		else:
			epl = self.rpc_endpoints
		fn = epl.pop(fn)
		if fn.is_alert != _alert:
			raise RuntimeError("register/unregister alert: %s/%s" % (fn.is_alert,_alert))
		if _async:
			if _alert:
				return self.conn.unregister_alert(fn)
			else:
				return self.conn.unregister_rpc(fn)

	def unregister_alert(self, fn):
		return self.unregister_rpc(fn, _alert=True)

	def unregister_rpc_async(self, fn):
		return self.unregister_rpc(fn, _async=True)
	unregister_rpc_sync = sync_maker("unregister_rpc_async")
	unregister_rpc_gevent = gevent_maker("unregister_rpc_async")

	def unregister_alert_async(self, fn):
		return self.unregister_rpc(fn, _async=True, _alert=True)
	unregister_alert_sync = sync_maker("unregister_alert_async")
	unregister_alert_gevent = gevent_maker("unregister_alert_async")

	def _alert_ping(self,msg):
		if isinstance(msg,Mapping):
			if msg.get('app',self.app) != self.app:
				return
			if msg.get('uuid',self.uuid) != self.uuid:
				return
		return dict(
			app=self.app,
			uuid=self.uuid,
			args=self.args,
			)

	def _reply_ping(self,msg):
		return dict(
			app=self.app,
			args=self.args,
			uuid=self.uuid,
			rpc=list(self.rpc_endpoints.keys()),
			alert=list(self.alert_endpoints.keys()),
			)
		
	## cleanup, less interesting (hopefully)

	def __del__(self):
		self.close(deleting=True)

	def close(self, deleting=False):
		self._kill_conn(deleting=deleting)

	def _kill_conn(self, deleting=False):
		c,self.conn = self.conn,None
		if c: # pragma: no cover
			try:
				c._kill()
			except Exception:
				if not deleting:
					logger.exception("closing connection")

	@asyncio.coroutine
	def _create_conn(self):
		from .conn import Connection
		conn = Connection(self)
		yield from conn.connect()
		for d in self.rpc_endpoints.values():
			yield from conn.register_rpc(d)
		for d in self.alert_endpoints.values():
			yield from conn.register_alert(d)
		self.conn = conn




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

# generic test setup

from yaml import safe_load

import logging,sys,os
logger = logging.getLogger("tests")

def test_init(who):
	if os.environ.get("TRACE","0") == '1':
		level = logging.DEBUG
	else:
		level = logging.WARN

	logger = logging.getLogger(who)
	logging.basicConfig(stream=sys.stderr,level=level)

	return logger

def load_cfg(cfg):
	"""load a config file"""
	global cfgpath
	if os.path.exists(cfg):
		pass
	elif os.path.exists(os.path.join("tests",cfg)):
		cfg = os.path.join("tests",cfg)
	elif os.path.exists(os.path.join(os.pardir,cfg)):
		cfg = os.path.join(os.pardir,cfg)
	else:
		raise RuntimeError("Config file '%s' not found" % (cfg,))

	cfgpath = cfg
	with open(cfg) as f:
		cfg = safe_load(f)

	return cfg



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

##
## Configuration: look up, in order:
## yaml_cfg.config
## etcd.specific.APP.config
## etcd.config
## 

import asyncio
from ..util import attrdict

import logging
logger = logging.getLogger(__name__)

# Calling conventions for RPC-registered procedures
CC_MSG="_msg" # pass the whole message (default)
CC_DATA="_data" # pass the data element
CC_DICT="_dict" # assume data is a dict and apply it

DEFAULT_CONFIG=dict(
	amqp=dict(
		server=dict(
			host='localhost',
			login='guest',
			password='guest',
			virtualhost='/qbroker',
			ssl=False,
			connect_timeout=10,
		),
		exchanges=dict(	  # all are persistent
			alert='alert', # topic: broadcast messages, may elicit multiple replies
			rpc='rpc',	 # topic: RPC requests, will trigger exactly one reply
			reply='reply', # direct: all replies go here
			dead='dead',   # fanout: dead messages (TTL expires, reject, RPC/alert unrouteable, …)
		),
		queues=dict(
			alert='alert_',# plus the unit UUID. Nonpersistent.
			rpc='rpc_',	# plus the command name. Persistent.
			reply='reply_',# plus the unit UUID
			dead='dead_',   # no add-on. Persistent. No TTL here!
		),
		handlers=dict(
			dead=False, # add a handler for dead messages
		),
		ttl=dict(
			rpc=10,
		),
		timeout=dict(
			rpc=15,
			poll=30,
		),
	))

@asyncio.coroutine
def make_unit(*a,**kw):
	"""\
		Create and start a QBroker unit.

		See qbroker.unit.Unit for parameters.
		"""
	u = Unit(*a,**kw)
	yield from u.start()
	return u

from .unit import Unit




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
## This file is part of QBroker, an easy to use RPC and broadcast
## client+server using AMQP.
##
## QBroker is Copyright © 2016 by Matthias Urlichs <matthias@urlichs.de>,
## it is licensed under the GPLv3. See the file `README.rst` for details,
## including optimistic statements by the author.
##
## This paragraph is auto-generated and may self-destruct at any time,
## courtesy of "make update". The original is in ‘utils/_boilerplate.py’.
## Thus, please do not remove the next line, or insert any blank lines.
##BP

import qbroker
qbroker.setup(sync=True,gevent=True)
from setuptools import setup, find_packages
import xml.sax.saxutils
from os.path import join
import sys
import os
from setuptools.command.test import test as TestCommand

def get_version(fname='qbroker/__init__.py'):
    with open(fname) as f:
        for line in f:
            if line.startswith('__VERSION__'):
                return eval(line.split('=')[-1])

def compile_po(path):
    from msgfmt import Msgfmt
    for language in os.listdir(path):
        lc_path = join(path, language, 'LC_MESSAGES')
        if os.path.isdir(lc_path):
            for domain_file in os.listdir(lc_path):
                if domain_file.endswith('.po'):
                    file_path = join(lc_path, domain_file)
                    mo_file = join(lc_path, '%s.mo' % domain_file[:-3])
                    mo_content = Msgfmt(file_path, name=file_path).get()
                    mo = open(mo_file, 'wb')
                    mo.write(mo_content)
                    mo.close()

def read(filename):
    text = open(filename,'r').read()
    return xml.sax.saxutils.escape(text)

class PyTest(TestCommand):
    #user_options = [('pytest-args=', 'a', "Arguments to pass to py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ['--assert=plain']

    def run_tests(self):
        #import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)

long_description = read('README.rst')

REQUIREMENTS = [i.strip() for i in open("requirements.txt").readlines()
                                if not i.startswith("http")]

setup(name='QBroker',
      license='GPLv3+',
      version='.'.join(str(x) for x in get_version()),
      description='QBroker is a minimal async wrapper for RPC via AMQP',
      long_description=long_description,
      author='Matthias Urlichs',
      author_email='matthias@urlichs.de',
      url='https://github.com/M-o-a-T/qbroker',
      install_requires=REQUIREMENTS,
      packages=find_packages(exclude=('tests',)),
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
          'Natural Language :: English',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Topic :: Software Development :: Libraries :: Application Frameworks',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: Utilities',
      ],
      zip_safe=True,
      cmdclass = {'test': PyTest},


      )




# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division, unicode_literals
##
##  This file is part of MoaT, the Master of all Things.
##
##  MoaT is Copyright © 2007-2015 by Matthias Urlichs <matthias@urlichs.de>,
##  it is licensed under the GPLv3. See the file `README.rst` for details,
##  including optimistic statements by the author.
##
##  This program is free software: you can redistribute it and/or modify
##  it under the terms of the GNU General Public License as published by
##  the Free Software Foundation, either version 3 of the License, or
##  (at your option) any later version.
##
##  This program is distributed in the hope that it will be useful,
##  but WITHOUT ANY WARRANTY; without even the implied warranty of
##  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##  GNU General Public License (included; see the file LICENSE)
##  for more details.
##
##  This header is auto-generated and may self-destruct at any time,
##  courtesy of "make update". The original is in ‘scripts/_boilerplate.py’.
##  Thus, do not remove the next line, or insert any blank lines above.
##BP

"""Collect codec info"""

from importlib import import_module
import os
import pkgutil

import logging
logger = logging.getLogger(__name__)

codecs = {}
default_codec = None
DEFAULT="DEFAULT"
def get_codec(name):
	global default_codec

	if not codecs:
		def _check(m):
			global default_codec
			try:
				if isinstance(m,str):
					m = import_module(m)
			except ImportError as ex:
				raise ImportError(m) from ex # pragma: no cover
				# not going to ship a broken file for testing this
			else:
				try:
					codecs[m.CODEC] = m
				except AttributeError:
					logger.warn("Not a codec")
					pass
				else:
					if getattr(m,'DEFAULT',False):
						default_codec = m

		for a,b,c in pkgutil.walk_packages((os.path.dirname(__file__),), __package__+'.'):
			_check(b)

	if name == DEFAULT:
		return default_codec
		
	return codecs[name]




#!/usr/bin/python
# -*- coding: utf-8 -*-

from __future__ import print_function,absolute_import
import sys
from time import mktime
from json.encoder import JSONEncoder
from json.decoder import JSONDecoder
from qbroker.util import attrdict, TZ,UTC, format_dt
import datetime as dt
from collections.abc import Mapping

CODEC = "application/binary"

DEBUG=False

def encode(data):
	if isinstance(data,bytes):
		data = data.decode('utf-8')
	return data

def decode(data):
	if isinstance(data,str):
		data = data.encode('utf-8')
	return data

